# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import sys

if sys.version_info >= (3, 11):
    from tomllib import load as toml_load
else:
    try:
        from tomli import load as toml_load
    except ImportError:
        from setuptools.extern.tomli import load as toml_load

from pbr.hooks import metadata as metadata_hooks
from pbr.setupcfg import split_multiline


def pbr(dist):
    """Inject dynamic config for PEP 517 / pyproject.toml-only builds.

    This is the setuptools.finalize_distribution_options hook. It handles
    projects that use only pyproject.toml with no setup.py (and therefore
    never trigger the distutils.setup_keywords handler in setupcfg.py).

    When setup.py with pbr=True is present, that handler sets
    _pbr_initialized before this hook runs, so we skip to avoid
    double-injection.
    """
    if hasattr(dist, '_pbr_initialized'):
        return

    try:
        with open("pyproject.toml", "rb") as f:
            pyproject = toml_load(f)
    except FileNotFoundError:
        return

    # Only activate for projects that have explicitly chosen pbr as their
    # build backend. Checking this prevents us from injecting into projects
    # that merely have pbr as a runtime dependency but use a different backend.
    build_backend = pyproject.get("build-system", {}).get("build-backend", "")
    if not build_backend.startswith("pbr"):
        return

    project = pyproject.get("project", {})
    dynamic = project.get("dynamic", [])

    name = dist.metadata.name or project.get("name")
    if not name:
        return

    dist._pbr_initialized = True

    config = {'metadata': {'name': name}}
    metadata_hooks.MetadataConfig(config).run()

    meta = config.get('metadata', {})
    if 'version' in meta and 'version' in dynamic:
        dist.metadata.version = meta['version']
    if 'requires_dist' in meta and 'dependencies' in dynamic:
        dist.install_requires = split_multiline(meta['requires_dist'])

    if 'sdist' not in dist.cmdclass:
        from pbr._compat.commands import LocalSDist

        dist.cmdclass['sdist'] = LocalSDist



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import
from __future__ import print_function

import os.path
import shlex
import sys

import fixtures
import packaging.utils
import testtools
import textwrap

from pbr._compat.five import ConfigParser
from pbr.tests import fixtures as pbr_fixtures
from pbr.tests.functional import base
from pbr.tests import util

PIPFLAGS = shlex.split(os.environ.get('PIPFLAGS', ''))
PIPVERSION = os.environ.get('PIPVERSION', 'pip')
PBRVERSION = os.environ.get('PBRVERSION', 'pbr')
REPODIR = os.environ.get('REPODIR', '')
WHEELHOUSE = os.environ.get('WHEELHOUSE', '')
PIP_CMD = ['-m', 'pip'] + PIPFLAGS + ['install', '-f', WHEELHOUSE]
PROJECTS = shlex.split(os.environ.get('PROJECTS', ''))


def all_projects():
    if os.environ.get('PBR_INTEGRATION', None) != '1':
        return

    if not REPODIR:
        return

    yielded = 0

    # Future: make this path parameterisable.
    excludes = set(['tempest', 'requirements'])
    for name in PROJECTS:
        name = name.strip()
        short_name = name.split('/')[-1]
        try:
            with open(
                os.path.join(REPODIR, short_name, 'setup.py'), 'rt'
            ) as f:
                if 'pbr' not in f.read():
                    continue
        except IOError:
            continue
        if short_name in excludes:
            continue
        yielded += 1
        yield (short_name, {'name': name, 'short_name': short_name})

    if not yielded:
        raise Exception(
            'no projects found: is PROJECTS set, and do the paths exist'
        )


class TestIntegration(base.BaseTestCase):

    scenarios = list(all_projects())

    def setUp(self):
        # Integration tests need a higher default - big repos can be slow to
        # clone, particularly under guest load.
        env = fixtures.EnvironmentVariable(
            'OS_TEST_TIMEOUT', os.environ.get('OS_TEST_TIMEOUT', '600')
        )
        with env:
            super(TestIntegration, self).setUp()
        util.config_git()

    @testtools.skipUnless(
        os.environ.get('PBR_INTEGRATION', None) == '1',
        'integration tests not enabled',
    )
    def test_integration(self):
        # Test that we can:
        # - run sdist from the repo in a venv
        # - install the resulting tarball in a new venv
        # - pip install the repo
        # - pip install -e the repo
        # We don't break these into separate tests because we'd need separate
        # source dirs to isolate from side effects of running pip, and the
        # overheads of setup would start to beat the benefits of parallelism.
        path = os.path.join(REPODIR, self.short_name)
        setup_cfg = os.path.join(path, 'setup.cfg')
        project_name = packaging.utils.canonicalize_name(
            self.short_name
        ).lower()
        # These projects should all have setup.cfg files but we'll be careful
        if os.path.exists(setup_cfg):
            config = ConfigParser()
            config.read(setup_cfg)
            if config.has_section('metadata'):
                raw_name = config.get(
                    'metadata', 'name', fallback='notapackagename'
                )
                # Technically we should really only need to use the raw
                # name because all our projects should be good and use
                # normalized names but they don't...
                project_name = packaging.utils.canonicalize_name(
                    raw_name
                ).lower()
        constraints = os.path.join(
            REPODIR, 'requirements', 'upper-constraints.txt'
        )
        tmp_constraints = os.path.join(
            self.useFixture(fixtures.TempDir()).path, 'upper-constraints.txt'
        )
        # We need to filter out the package we are installing to avoid
        # conflicts with the constraints.
        with open(constraints, 'r') as src:
            with open(tmp_constraints, 'w') as dest:
                for line in src:
                    constraint = line.split('===')[0]
                    constraint = packaging.utils.canonicalize_name(
                        constraint
                    ).lower()
                    if project_name != constraint:
                        dest.write(line)
        pip_cmd = PIP_CMD + ['-c', tmp_constraints]

        with pbr_fixtures.Venv(
            'sdist',
            modules=['pip', 'wheel', 'setuptools<80', PBRVERSION],
            pip_cmd=PIP_CMD,
        ) as venv:
            python = venv.python
            pbr_fixtures.CapturedSubprocess(
                'sdist',
                [python, 'setup.py', 'sdist'],
                cwd=path,
                test_case=self,
            )

        with pbr_fixtures.Venv(
            'tarball',
            modules=['pip', 'wheel', 'setuptools<80', PBRVERSION],
            pip_cmd=PIP_CMD,
        ) as venv:
            python = venv.python
            filename = os.path.join(
                path, 'dist', os.listdir(os.path.join(path, 'dist'))[0]
            )
            pbr_fixtures.CapturedSubprocess(
                'tarball',
                [python] + pip_cmd + [filename],
                test_case=self,
            )

        with pbr_fixtures.Venv(
            'install-git',
            modules=['pip', 'wheel', 'setuptools<80', PBRVERSION],
            pip_cmd=PIP_CMD,
        ) as venv:
            root = venv.path
            python = venv.python
            pbr_fixtures.CapturedSubprocess(
                'install-git',
                [python] + pip_cmd + ['git+file://' + path],
                test_case=self,
            )
            if self.short_name == 'nova':
                found = False
                for _, _, filenames in os.walk(root):
                    if 'alembic.ini' in filenames:
                        found = True
                self.assertTrue(found)

        with pbr_fixtures.Venv(
            'install-editable',
            modules=['pip', 'wheel', 'setuptools<80', PBRVERSION],
            pip_cmd=PIP_CMD,
        ) as venv:
            python = venv.python
            pbr_fixtures.CapturedSubprocess(
                'install-editable',
                [python] + pip_cmd + ['-e', path],
                test_case=self,
            )


class TestInstallWithoutPbr(base.BaseTestCase):

    # TODO(clarkb) This test should be reimagined with modern packaging tools
    # and expectations.
    @testtools.skipUnless(
        os.environ.get('PBR_INTEGRATION', None) == '1',
        'integration tests not enabled',
    )
    def test_install_without_pbr(self):
        # Test easy-install of a thing that depends on a thing using pbr
        tempdir = self.useFixture(fixtures.TempDir()).path
        # A directory containing sdists of the things we're going to depend on
        # in using-package.
        dist_dir = os.path.join(tempdir, 'distdir')
        os.mkdir(dist_dir)
        self._run_cmd(
            sys.executable,
            ('setup.py', 'sdist', '-d', dist_dir),
            allow_fail=False,
            cwd=pbr_fixtures.PBR_ROOT,
        )
        # testpkg - this requires a pbr-using package
        test_pkg_dir = os.path.join(tempdir, 'testpkg')
        os.mkdir(test_pkg_dir)
        pkgs = {
            'pkgTest': {
                # override setup.py, setup.cfg since we don't want pbr used for
                # this package
                'setup.py': textwrap.dedent(
                    """\
                    #!/usr/bin/env python
                    import setuptools
                    setuptools.setup(
                        name = 'pkgTest',
                        # TODO(clarkb) should we use a random prefix to
                        # avoid collisions?
                        install_requires = ['pkgReq'],
                    )
                    """
                ),
                'setup.cfg': textwrap.dedent(
                    """\
                    [easy_install]
                    find_links = %s
                    """
                    % dist_dir
                ),
            },
            # We don't need to use PBRVERSION here because we precreate the
            # pbr sdist and point to it with find_links.
            'pkgReq': {
                # ...but we use the standard setup.py, setup.cfg provided by
                # the fixture since we do want pbr here, in the dependency
                'requirements.txt': textwrap.dedent(
                    """\
                    pbr
                    """
                ),
                'pkgReq/__init__.py': "",
                'pkgReq/__main__.py': textwrap.dedent(
                    """\
                    print("FakeTest loaded and ran")
                    """
                ),
            },
        }
        pkg_dirs = self.useFixture(pbr_fixtures.Packages(pkgs)).package_dirs
        test_pkg_dir = pkg_dirs['pkgTest']
        req_pkg_dir = pkg_dirs['pkgReq']

        self._run_cmd(
            sys.executable,
            ('setup.py', 'sdist', '-d', dist_dir),
            allow_fail=False,
            cwd=req_pkg_dir,
        )
        # A venv to test within
        # We install setuptools because we rely on setup.py below.
        # FIXME(stephenfin): We should not need to pin setuptools
        # https://github.com/pypa/setuptools/commit/ef4cd2960d75f2d49f40f5495347523be62d20e5
        with pbr_fixtures.Venv(
            'nopbr', ['pip', 'wheel', 'setuptools<80']
        ) as venv:
            python = venv.python
            # Install both packages
            pbr_fixtures.CapturedSubprocess(
                'nopbr',
                [python] + ['setup.py', 'install'],
                cwd=test_pkg_dir,
                test_case=self,
            )
            # Execute code that should only be present if the install worked.
            pbr_fixtures.CapturedSubprocess(
                'nopbr',
                [python] + ['-m', 'pkgReq'],
                cwd=test_pkg_dir,
                test_case=self,
            )
            pbr_cmd = os.path.join(venv.path, 'bin', 'pbr')
            pbr_fixtures.CapturedSubprocess(
                'nopbr',
                [pbr_cmd] + ['freeze'],
                cwd=test_pkg_dir,
                test_case=self,
            )


# Handle various compatability issues with pip and setuptools versions against
# python3 versions. Unfortunately python3.12 in particular isn't very backward
# compatible with pip and setuptools.
# TODO(clarkb) add other distros like EL9 and EL10
if sys.version_info[0:3] < (3, 10, 0):
    lts_scenarios = [
        ('Bionic', {'modules': ['pip==9.0.1', 'setuptools==39.0.1']}),
        ('Stretch', {'modules': ['pip==9.0.1', 'setuptools==33.1.1']}),
        ('EL8', {'modules': ['pip==9.0.3', 'setuptools==39.2.0']}),
        ('Buster', {'modules': ['pip==18.1', 'setuptools==40.8.0']}),
        ('Focal', {'modules': ['pip==20.0.2', 'setuptools==45.2.0']}),
    ]
elif sys.version_info[0:3] < (3, 12, 0):
    lts_scenarios = [
        ('Bullseye', {'modules': ['pip==20.3.4', 'setuptools==52.0.0']}),
        ('Bookworm', {'modules': ['pip==23.0.1', 'setuptools==66.1.1']}),
        ('Focal', {'modules': ['pip==20.0.2', 'setuptools==45.2.0']}),
        ('Jammy', {'modules': ['pip==22.0.2', 'setuptools==59.6.0']}),
    ]
else:
    lts_scenarios = [
        ('Noble', {'modules': ['pip==24.0.0', 'setuptools==68.1.2']}),
    ]


class TestMarkersPip(base.BaseTestCase):

    scenarios = [
        ('pip-latest', {'modules': ['pip', 'setuptools']})
    ] + lts_scenarios

    @testtools.skipUnless(
        os.environ.get('PBR_INTEGRATION', None) == '1',
        'integration tests not enabled',
    )
    def test_pip_versions(self):
        pkgs = {
            'test_markers': {
                'requirements.txt': textwrap.dedent(
                    """\
                    pkg_a; python_version=='1.2'
                    pkg_b; python_version!='1.2'
                """
                )
            },
            'pkg_a': {},
            'pkg_b': {},
        }
        pkg_dirs = self.useFixture(pbr_fixtures.Packages(pkgs)).package_dirs
        temp_dir = self.useFixture(fixtures.TempDir()).path
        repo_dir = os.path.join(temp_dir, 'repo')
        with pbr_fixtures.Venv('markers') as venv:
            bin_python = venv.python
            os.mkdir(repo_dir)
            for module in self.modules:
                pbr_fixtures.CapturedSubprocess(
                    'pip-version',
                    [
                        bin_python,
                        '-m',
                        'pip',
                        'install',
                        '--upgrade',
                        module,
                    ],
                    cwd=venv.path,
                    test_case=self,
                )
            # TODO(clarkb) do we need to install PBR from source here to avoid
            # using the latest release?
            for pkg in pkg_dirs:
                self._run_cmd(
                    bin_python,
                    ['setup.py', 'sdist', '-d', repo_dir],
                    cwd=pkg_dirs[pkg],
                    allow_fail=False,
                )
            self._run_cmd(
                bin_python,
                [
                    '-m',
                    'pip',
                    'install',
                    '--no-index',
                    # With --no-index above we can't install deps for isolated
                    # builds. Just rely on the existing installs for now.
                    '--no-build-isolation',
                    '-f',
                    repo_dir,
                    'test_markers',
                ],
                cwd=venv.path,
                allow_fail=False,
            )
            pkgs = self._run_cmd(
                bin_python,
                ['-m', 'pip', 'freeze'],
                cwd=venv.path,
                allow_fail=False,
            )[0]
            # Depending on the version of pip/setuptools etc the name of the
            # installed package may be noramlized to 'pkg-b'. As of March 2024
            # 'pkg_b' is what we get and previously 'pkg-b' was the result.
            self.assertTrue('pkg_b' in pkgs or 'pkg-b' in pkgs)


class TestLTSSupport(base.BaseTestCase):

    scenarios = lts_scenarios

    @testtools.skipUnless(
        os.environ.get('PBR_INTEGRATION', None) == '1',
        'integration tests not enabled',
    )
    def test_lts_venv_default_versions(self):
        with pbr_fixtures.Venv('setuptools', modules=self.modules) as venv:
            bin_python = venv.python
            pbr = 'file://%s#egg=pbr' % pbr_fixtures.PBR_ROOT
            # Installing PBR is a reasonable indication that we are not broken
            # on this particular combination of setuptools and pip.
            pbr_fixtures.CapturedSubprocess(
                'lts-support',
                [bin_python, '-m', 'pip', 'install', pbr],
                cwd=venv.path,
                test_case=self,
            )



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import sys


def pbr(dist):
    """The ``setuptools.finalize_distribution_options`` entry point.

    setuptools runs *every* registered finalize_distribution_options hook
    whenever *any* ``Distribution`` is constructed -- including by unrelated
    tooling such as virtualenv, which builds a throwaway distutils
    ``Distribution`` purely to probe install paths. Resolving this entry
    point therefore imports its target module at arbitrary times and from
    arbitrary working directories.

    That is why this shim lives in the (always already-imported) ``pbr``
    package rather than in ``pbr.pyprojecttoml``. Under Python 2 develop
    installs ``pbr`` is found via a relative ``sys.path`` entry, so
    ``pbr.__path__`` is relative (``['pbr']``). Importing a not-yet-loaded
    submodule after the process has changed directory then fails with
    ``ImportError: No module named pyprojecttoml``. Anchoring the hook here
    -- and only importing the implementation on the Python versions that can
    actually build from a pyproject.toml -- avoids that fragile late import.
    """
    if sys.version_info < (3, 7):
        # pyproject.toml-only builds require tomllib/tomli, which we do not
        # support on Python 2.7 or 3.6. There is nothing to do here, and
        # crucially we must not import pbr.pyprojecttoml in this case.
        return

    from pbr import pyprojecttoml

    pyprojecttoml.pbr(dist)


# Ensure this hook runs after setuptools' built-in 'keywords' hook
pbr.order = 1



# Copyright (c) 2013 New Dream Network, LLC (DreamHost)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Copyright (C) 2013 Association of Universities for Research in Astronomy
#                    (AURA)
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#     1. Redistributions of source code must retain the above copyright
#        notice, this list of conditions and the following disclaimer.
#
#     2. Redistributions in binary form must reproduce the above
#        copyright notice, this list of conditions and the following
#        disclaimer in the documentation and/or other materials provided
#        with the distribution.
#
#     3. The name of AURA and its representatives may not be used to
#        endorse or promote products derived from this software without
#        specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY AURA ``AS IS'' AND ANY EXPRESS OR IMPLIED
# WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL AURA BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS

import os
import textwrap

import packaging.markers
import packaging.requirements

from pbr._compat.five import string_type
from pbr.tests import fixtures as pbr_fixtures
from pbr.tests.functional import base


# Taken from setuptools
#
# https://github.com/pypa/setuptools/blob/v44.1.1/pkg_resources/__init__.py#L2378-L2389
def yield_lines(strs):
    """Yield non-empty/non-comment lines of a string or sequence"""
    if isinstance(strs, string_type):
        for s in strs.splitlines():
            s = s.strip()
            # skip blank lines/comments
            if s and not s.startswith('#'):
                yield s
    else:
        for ss in strs:
            for s in yield_lines(ss):
                yield s


# Taken from setuptools
#
# https://github.com/pypa/setuptools/blob/v44.1.1/pkg_resources/__init__.py#L3189-L3212
def split_sections(s):
    """Split a string or iterable thereof into (section, content) pairs

    Each ``section`` is a stripped version of the section header ("[section]")
    and each ``content`` is a list of stripped lines excluding blank lines and
    comment-only lines.  If there are any such lines before the first section
    header, they're returned in a first ``section`` of ``None``.
    """
    section = None
    content = []
    for line in yield_lines(s):
        if line.startswith("["):
            if line.endswith("]"):
                if section or content:
                    yield section, content
                section = line[1:-1].strip()
                content = []
            else:
                raise ValueError("Invalid section heading", line)
        else:
            content.append(line)

    # wrap up last segment
    yield section, content


class TestRequirementParsing(base.BaseTestCase):

    @staticmethod
    def _normalize_section_key(key):
        """Normalize a requires.txt section key for comparison.

        setuptools normalizes markers when writing requires.txt from
        PEP 508 inline markers. This helper strips differences like
        parentheses, quoting style, and whitespace so we can compare
        keys regardless of setuptools version.
        """
        if key is None:
            return key
        if ':' in key:
            name, marker = key.split(':', 1)
            marker = marker.strip().strip('()')
            marker = str(packaging.markers.Marker(marker))
            return '%s:%s' % (name, marker) if name else ':%s' % marker
        return key

    def test_requirement_parsing(self):
        pkgs = {
            'test_reqparse': {
                'requirements.txt': textwrap.dedent(
                    """\
                        bar
                        quux<1.0; python_version=='2.6'
                        requests-aws>=0.1.4    # BSD License (3 clause)
                        Routes>=1.12.3,!=2.0,!=2.1;python_version=='2.7'
                        requests-kerberos>=0.6;python_version=='2.7' # MIT
                    """
                ),
                'setup.cfg': textwrap.dedent(
                    """\
                        [metadata]
                        name = test_reqparse

                        [extras]
                        test =
                            foo
                            baz>3.2 :python_version=='2.7' # MIT
                            bar>3.3 :python_version=='2.7' # MIT # Apache
                    """
                ),
            },
        }
        pkg_dirs = self.useFixture(pbr_fixtures.Packages(pkgs)).package_dirs
        pkg_dir = pkg_dirs['test_reqparse']
        # pkg_resources.split_sections uses None as the title of an
        # anonymous section instead of the empty string. Weird.
        expected_requirements = {
            None: ['bar', 'requests-aws>=0.1.4'],
            ":(python_version=='2.6')": ['quux<1.0'],
            ":(python_version=='2.7')": [
                'Routes!=2.0,!=2.1,>=1.12.3',
                'requests-kerberos>=0.6',
            ],
            'test': ['foo'],
            "test:(python_version=='2.7')": ['baz>3.2', 'bar>3.3'],
        }
        # Two things are tested by this
        # 1) pbr properly parses markers from requirements.txt and setup.cfg
        # 2) bdist_wheel causes pbr to not evaluate markers
        with pbr_fixtures.Venv('reqParse') as venv:
            self._run_cmd(
                venv.python,
                ('setup.py', 'bdist_wheel'),
                allow_fail=False,
                cwd=pkg_dir,
            )
            egg_info = os.path.join(pkg_dir, 'test_reqparse.egg-info')

            requires_txt = os.path.join(egg_info, 'requires.txt')
            with open(requires_txt, 'rt') as requires:
                generated_requirements = dict(split_sections(requires))

            # Normalize section keys: setuptools may normalize markers
            # differently depending on how they are passed (legacy
            # extras_require key vs PEP 508 inline).
            normalized_generated = {
                self._normalize_section_key(k): v
                for k, v in generated_requirements.items()
            }

            # NOTE(dhellmann): We have to spell out the comparison because
            # the rendering for version specifiers in a range is not
            # consistent across versions of setuptools.

            for section, expected in expected_requirements.items():
                norm_section = self._normalize_section_key(section)
                # We wrap in str since we need packaging 22.0.0 or later to do
                # comparisons [1] and that doesn't support Python 2.7, 3.6
                #
                # https://github.com/pypa/packaging/commit/aebc072a06925cc0004b031e6b6f3028e5e2e686
                # https://pypi.org/project/packaging/22.0/
                exp_parsed = [
                    str(packaging.requirements.Requirement(s))
                    for s in expected
                ]
                gen_parsed = [
                    str(packaging.requirements.Requirement(s))
                    for s in normalized_generated[norm_section]
                ]
                self.assertEqual(exp_parsed, gen_parsed)
